'use strict';
const request = require('request');
const rperrors = require('request-promise/errors');
const rp = require('request-promise');
const keytar = require('keytar');
const path = require('path');
const crypto = require('@trust/webcrypto')
const fs = require('fs');
const sha256 = require('sha256');

const ACCOUNTSLABEL = 'accounts'
const SystemCollection = "buckets";
const SMTTreeDepth = 160;

/* TransactionType Constants */
const TransactionTransfer = 1
const TransactionRegisterNode = 2
const TransactionUpdateNode = 3
const TransactionBandwidthCheck = 4
const TransactionSetQuota = 5
const TransactionSetKey = 6
const TransactionDeleteKey = 7
const TransactionSetName = 8
const TransactionMutate = 9

/* BucketType Constants */
const BucketCloudstore = 0
const BucketNoSQL = 1
const BucketSQL = 2
const DefaultMaxBucketSize = 1000000

/* SQLREQUEST OPTION Constants */
const RT_PUT = 'Put'
const RT_GET = "Get"
const RT_DELETE = "Delete"
const RT_QUERY = "Query"
const RT_SCAN = "Scan"

const APPNAME = 'Wolk';
const emptyHash = "0x0000000000000000000000000000000000000000000000000000000000000000"; //new Uint8Array(32);
const DEFAULTPROVIDER = "https://cloud.wolk.com"

const DEFAULT_DEVELOPER_TRUST_LEVEL = 1 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time
const DEFAULT_USER_TRUST_LEVEL = 1 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time


class WOLK {
    constructor() {
      console.log("[wolkjs/browser] constructor")
      this.developerTrustLevel = DEFAULT_DEVELOPER_TRUST_LEVEL
      this.userTrustLevel = DEFAULT_USER_TRUST_LEVEL
      this.setWolkTrustLevel()
    }

    async init() {
      console.log("[wolkjs/browser] init called")
      await this.getProvider()
      this.ecdsaPublicKeyData = '';
      this.ecdsaPrivateKeyData = '';
      this.rsaPrivateKeyData = '';
      this.rsaPublicKeyData = '';
      this.activeAccount = '';
      this.activeAccountName = ''
      let accts = await this.getAccounts()
      //console.log("[wolkjs/browser] init accounts: " + JSON.stringify(accts))
      for(var index = 0; index<this.localAccounts.length; index++) {
        var currentAccount = this.localAccounts[index]
        if ( currentAccount.isActive && this.activeAccountName != currentAccount.name ) {
          await this.setActiveAccount( currentAccount.name )
        }
      }
      /*
        names =
        [{
          name: "Rodney",
          keys: {
            privateKey: "AAA",
            friendsKey: "BBB",
            rsaPrivateKey:"CCC",
            rsaPublicKey: "DDD"
          },
          isLoggedIn: true,
          lastLogin: ???
        }, {
          name:"Shaw",
          keys: {
            privatKey: "VVV",
            friendsKey: "XXX",
            rsaPrivateKey:"YYY",
            rsaPublicKey: "ZZZ"
          },
          isLoggedIn: false,
        }];
      */
      //TODO: Add Provider to names???
    }

    async login( name, password ) {
      let accts = await this.getAccounts()
      //console.log("Accounts: " + JSON.stringify(accts))
      for(var index = 0; index<this.localAccounts.length; index++) {
        var currentAccount = this.localAccounts[index]
        console.log("Checking account: " + currentAccount.name + " vs " + name)
        if ( currentAccount.name == name ) {
          console.log("[wolkjs/browser] Logging in as: " + currentAccount.name)
          return this.setActiveAccount( currentAccount.name );          
        }
      }
    }

    setDeveloperTrustLevel( level ) {
      this.developerTrustLevel = level
    }

    setUserTrustLevel( level ) {
      this.userTrustLevel = level
    }

    setWolkTrustLevel() {
      if( this.userTrustLevel > this.developerTrustLevel ) {
        this.wolkTrustLevel  = this.userTrustLevel
      } else {
        this.wolkTrustLevel  = this.developerTrustLevel
      }
    }

    determineProofValue() {
      let randProof = Math.random();
      if ( randProof <= this.wolkTrustLevel ) {
        return 1
      } else {
        return 0
      }
    }

    async setProvider(provider) {
      console.log("[index:setProvider] START - setting to: [" + provider + "]")
      return keytar.setPassword(APPNAME, 'provider', provider)
      .then( () => {
        this.provider = provider;
        return provider;
        console.log("Provider Set to: " + provider)
      })
      .catch( (err) => {
        console.log('Ooops!', err) // some kind of I/O error
      });
    }

    async getProvider() {
      console.log("[index:getProvider] START - getting from: [" + this.provider + "] for APPNAME " +APPNAME )
      if ( this.provider != undefined ) {
        return this.provider;
      }
      return keytar.getPassword(APPNAME, "provider")
      .then( (providerValue) => {
        if (providerValue == null) {
          console.log("WTF: why is it not pulling from keytar")
          this.provider = DEFAULTPROVIDER
        } else {
          console.log("[index:getProvider] returning provider: " + providerValue)
          this.provider = providerValue;
        }
        return this.provider
      })
      .catch( (err) => {
        console.log("ERROR: unable to retrieve provider")
      })
    }

    async accountExists(name) {
      await this.getAccounts()

      if( this.localAccounts.length == 0 ) {
        console.log("[wolkjs/browser] Account Does Not Exist:" + name + " - returning")
        return {};
      }
      for(var index = 0; index<this.localAccounts.length; index++) {
        var currentAccount = this.localAccounts[index]
        //console.log("Checking account: " + currentAccount.name + " vs " + name)
        if ( currentAccount.name == name ) {
          console.log("[wolkjs/browser] Account Exists: " + currentAccount.name)
          return currentAccount;
        }
      }
      console.log("[wolkjs/browser] Account Does Not Exist: " + currentAccount.name)
      return {};
    }

    async getActiveAccount() {
      await this.getAccounts()
      for(var index = 0; index<this.localAccounts.length; index++) {
        var currentAccount = this.localAccounts[index]
        if ( currentAccount.isActive ) {
          return currentAccount
        }
      }
    }    

    async loadActiveAccount() {
      await this.getAccounts()
      var found = false;
      for(var index = 0; index<this.localAccounts.length; index++) {
        var currentAccount = this.localAccounts[index]
        if ( currentAccount.isActive ) {
          this.setActiveAccount(currentAccount.name);
        }
      }
    }

    async getAccounts() {
      console.log("[beaker/wolkjs] getAccounts")
      return await keytar.getPassword(APPNAME, ACCOUNTSLABEL)
      .then( async (names) => {
        if (names == null) {
          console.log("WARNING: NO Accounts found -- returning []")
          this.localAccounts = [];
          return [];
        }
        console.log("GOT Accounts: ")
        // console.log(names)
        this.localAccounts = JSON.parse(names);
        return this.localAccounts
      })
      .catch( (err) => {
        console.log("ERROR: Can't get localNames - " + err)
        return [];
      });
    }

    async saveAccount( account ) {
      await this.getAccounts().catch( (err) => {
          console.log("ERROR: Can't get LocalAccounts - " + err);
      });
      //console.log("Retrieved localAccounts " + JSON.stringify(this.localAccounts));
      var found =false;
      for (var index = 0; index < this.localAccounts.length; ++index) {
        var currentAccount = this.localAccounts[index];
        if(currentAccount.name == account.name){
          console.log("WARNING: Name already exists - NOT Updating keys")
          //TODO: could attempt to retrieve from cloudstore to check that the key entered matches that??
          found = true;
          break;
        }
      }
      if ( !found ) {
        //console.log( " Adding account: " + JSON.stringify(account) )
        this.localAccounts.push( account )
      }
      console.log("saving account to keytar")
      return keytar.setPassword(APPNAME, ACCOUNTSLABEL, JSON.stringify(this.localAccounts))
      .then( () => {
        console.log("SUCCESS: updated localAccounts to include: " + account.name );
      })
      .catch( (err) => {
        console.log("ERROR: setting localName - " + err)
      })
    }

    async setActiveAccount( accountName ) {
      accountName = accountName.toLowerCase();
      if ( accountName == this.activeAccountName ) {
        console.log("[wolk/browser:setActiveAccount] account already active - returning")
        return
      }

      let foundAccount = await this.accountExists(accountName);
      //accountExists loads accounts locally
      if( Object.keys(foundAccount).length == 0 ) {
        console.log("[browser:loadAccount] cannot find: [" + accountName + "]")
        //TODO: throw error
        return;
      }

      console.log("Setting Active Account to: " + accountName)
      this.activeAccountName = foundAccount.name
      this.ecdsaPublicKeyData = foundAccount.keys.ecdsaPublicKeyData
      this.ecdsaPrivateKeyData = foundAccount.keys.ecdsaPrivateKeyData
      this.rsaPrivateKeyData = foundAccount.keys.rsaPrivateKeyData
      this.rsaPublicKeyData = foundAccount.keys.rsaPublicKeyData
      await crypto.subtle.importKey( "jwk", JSON.parse(this.ecdsaPrivateKeyData),  {name: "ECDSA",  namedCurve: "P-256" },  false,  ['sign'] )
      .then( (result) => {
        console.log("[wolkjs/browser:setActiveAccount] - Importing Key: " + JSON.stringify(result))
        console.log("[wolkjs/browser:setActiveAccount] - ")
        this.ecdsaKey = {}
        this.ecdsaKey.privateKey = result;
        this.ecdsaPublicKey = foundAccount.keys.ecdsaPublicKeyData
        for( var index =0; index < this.localAccounts.length; index++ ) {
          var currentAccount = this.localAccounts[index];
          if ( currentAccount.name == accountName ) {
            this.localAccounts[index].isActive = true;
          } else {
            this.localAccounts[index].isActive = false;
          }
        }
        console.log("[wolkjs/browser] Attempt to update Keytar")
        return keytar.setPassword(APPNAME, ACCOUNTSLABEL, JSON.stringify(this.localAccounts))
        .then( () => {
          console.log("SUCCESS: updated localAccounts to activate: " + accountName );
          return true;
        })
        .catch( (err) => {
          console.log("ERROR: setting localName - " + err)
          return false;
        })
      })
      .catch( (error) => { console.log("Error Getting ECDSA Private key: " + error ); });
      // TODO: symmetric keys
    }

    async createAccount(name, options, callback) {
      name = name.toLowerCase();
      console.log("createAccount:", name)
      // generate RSA + ECDSA Keys
      let rsaKey = await crypto.subtle.generateKey({name: "RSASSA-PKCS1-v1_5", modulusLength: 2048, publicExponent: new Uint8Array([0x01, 0x00, 0x01]), hash: {name: "SHA-256"}}, true, ["sign", "verify"]);
      let ecdsaKey = await crypto.subtle.generateKey( { name: "ECDSA", namedCurve: "P-256"  },  true, ["sign", "verify"] );
      this.rsaKey = rsaKey.privateKey
      this.ecdsaKey = ecdsaKey

      // export RSA + ECDSA Keys
      let rsaPublicKey = await crypto.subtle.exportKey("jwk", rsaKey.publicKey)
      let rsaPrivateKey = await crypto.subtle.exportKey("jwk", rsaKey.privateKey)
      let ecdsaPublicKey = await crypto.subtle.exportKey("jwk", ecdsaKey.publicKey)
      let ecdsaPrivateKey = await crypto.subtle.exportKey("jwk", ecdsaKey.privateKey)
      this.ecdsaPublicKeyData = JSON.stringify(ecdsaPublicKey);
      this.ecdsaPrivateKeyData = JSON.stringify(ecdsaPrivateKey);
      this.rsaPrivateKeyData = JSON.stringify(rsaPrivateKey);
      this.rsaPublicKeyData = JSON.stringify(rsaPublicKey);

      // save files
      let accountKeys = {
        ecdsaPublicKeyData: this.ecdsaPublicKeyData,
        ecdsaPrivateKeyData: this.ecdsaPrivateKeyData,
        rsaPrivateKeyData: this.rsaPrivateKeyData,
        rsaPublicKey: this.rsaPublicKeyData
      };

      let newAccount = {
        name: name,
        keys: accountKeys
      };
      return this.saveAccount(newAccount)
      .then( async () => {
        console.log("Saved keys")
        if ( this.activeAccountName == undefined ) {
          // set to Active
          await this.setActiveAccount( newAccount.name )
        }
        if ( typeof callback === 'function' ) {
          console.log("calling setname w/ callback => " + callback)
          return this.setName(name, JSON.stringify(rsaPublicKey), callback)
        } else {
          console.log("calling setname w/ promise | name: " + name + " | " + JSON.stringify(rsaPublicKey) + " | " + callback)
          return this.setName(name, JSON.stringify(rsaPublicKey), options)
            .then(([hash]) => {
              console.log("SUCC: setname resolved to hash: " + hash.toString('utf8'));
              return hash.toString('utf8');
             })
            .catch((err) => { console.log("ERROR: setname - ", err)});
        }
      })
      .catch((err) => {
        console.log("ERROR: saveaccount - ", err)
      });      
    }

    setName(name, rsaPublicKey, options, callback) {
      return this.post(name, {
        "name": name,
        "rsaPublicKey": Buffer.from(rsaPublicKey).toString('base64')
      }, options, callback);
    }

    // Blockchain operations
    getBlock(blockNumber, callback) {
      return this.get(joinPath('wolk', 'block', blockNumber.toString(10)), {}, callback)
              .catch( (err) => { throw err; });
    }
    getNode(nodeNumber, callback) {
      return this.get(joinPath('wolk', 'node', nodeNumber.toString(10)), {}, callback)
              .catch( (err) => { throw err; });
    }
    updateNode(nodeNumber, nodeInfo, callback) {
      return this.patch(joinPath('wolk', 'node', nodeNumber.toString(10)), nodeInfo, {}, callback);
    }
    getName(name, callback) {
      return this.get(joinPath('wolk', 'name', name), {}, callback)
              .catch( (err) => { throw err; });
    }
    getAccount(name, callback) {
      return this.get(joinPath('wolk', 'account', name), {}, callback)
              .catch( (err) => { throw err; });
    }
    getLatestBlockNumber(callback) {
      return this.get(joinPath('wolk', 'block', 'latest'), {}, callback)
              .catch( (err) => { throw err; });
    }
    getBalance(name_address, callback) {
      return this.get(joinPath('wolk', 'balance', name_address), {}, callback)
              .catch( (err) => { throw err; });
    }
    getTransaction(txhash, callback) {
      return this.get(joinPath('wolk', 'tx', txhash), {}, callback)
              .catch( (err) => { throw err; });
    }
    transfer(recipient, amount, callback) {
      return this.post(joinPath('wolk', 'transfer'), {
        "recipient": recipient,
        "amount": amount
      }, {}, callback);
    }

    //TODO: rather than pass / hdrs - consider storing alongside accounts etc ... and use from there
    // NoSQL
    createBucket(owner, collection, callback) {
      return this.post(joinPath(owner, collection), JSON.stringify({
        "name": collection,
        "bucketType": BucketCloudstore,
        "requesterPays": 0,
        "encryption": "none"
      }), {}, callback);
    }
    createCollection(owner, collection, hdrs, callback) {
      return this.post(joinPath(owner, collection), JSON.stringify({
        "name": collection,
        "bucketType": BucketNoSQL,
        "requesterPays": 0,
        "encryption": "none"
      }), hdrs, callback);
    }
    listCollections(owner, hdrs, callback) {
      hdrs.Proof = this.determineProofValue();
      return this.get(owner, hdrs, callback)
      .catch( (err) => { throw err; });
    }

    deleteCollection(owner, collection, callback) {
      return this.delete(joinPath(owner, collection), {}, callback);
    }
    scanCollection(owner, collection, hdrs, callback) {
      hdrs.Proof = this.determineProofValue();
      return this.get(joinPath(owner, collection), hdrs, callback)
        .catch( (err) => { throw err; });
    }
    setKey(owner, collection, key, value, hdrs, callback) {
      return this.put(joinPath(owner, collection, key), value, hdrs, callback);
    }
    getKey(owner, collection, key, hdrs, callback) {    
      if (Object.keys(hdrs).length && hdrs.hasOwnProperty("Proof")) {
        hdrs = hdrs;
      }
      else {
        hdrs.Proof = this.determineProofValue();
      }      
      return this.get(joinPath(owner, collection, key), hdrs, callback)
        .catch( (err) => { throw err; });
    }
    getChunk( chunkId, options, callback) {
      return this.get(joinPath(ProtocolName, PayloadChunk, chunkId), { "Proof": this.determineProofValue() }, callback)
            .catch( (err) => { throw err; });
    }
    deleteKey(owner, collection, key, callback) {
      return this.delete(joinPath(owner, collection, key), {}, callback);
    }

    // SQL operations
    createDatabase(owner, database, options, callback) {
      return this.post(joinPath(owner, database), JSON.stringify({
        "name": database,
        "bucketType": BucketSQL,
        "requesterPays": 0,
        "encryption": "none"
      }), options, {}, callback);
    }
    listDatabases(owner, callback) {
      return this.get(joinPath(owner), {
          "Proof": this.determineProofValue()
        }, callback)
        .catch( (err) => {
          console.log("ERR: listDatabases - " + err)
        });
    }
    deleteDatabase(owner, database, callback) {
      return this.delete(joinPath(owner, collection), {}, callback);
    }
    executeSQL(owner, database, sql, hdrs, callback) {
      return this.post(joinPath(owner, database, sql), sql, hdrs, callback);
    }
    /* TODO:
    async createTable(owner, database, table, schema, callback) { return this.post(this.provider + "/" + owner + "/" + database + "/" + table, schema, callback); }
    describeTable(owner, database, table, hdrs, callback) { return this.post(this.provider + "/" + owner + "/" + database + "/" + table, hdrs, callback); }
    async dropTable(owner, database, table, callback) { return this.get(this.provider + "/" + owner + "/" + database + "/" + table, callback); }
    listTables(owner, database, hdrs, callback) { return this.get(this.provider + "/" + owner + "/" + SystemCollection + "/SCAN", hdrs, callback); }
    */
    async get(path, hdrs, callback) {
      return this.wrequest("GET", path, null, hdrs, callback);
    }

    async post(path, body, hdrs, callback) {
      return this.wrequest("POST", path, body, hdrs, callback);
    }

    async put(path, body, hdrs, callback) {
      return this.wrequest("PUT", path, body, hdrs, callback);
    }

    async delete(path, hdrs, callback) {
      return this.wrequest("DELETE", path, null, hdrs, callback);
    }

    async head(wolkUrl, hdrs, callback) {
      if (! isWolkUrl(wolkUrl)) {   throw new Error("Invalid WolkUrl: " + wolkUrl)  }
      var payload = wolkUrl.substr(6, wolkUrl.length-6);
      let wolkUrlObject = parseWolkUrl(wolkUrl);
      let that = this
      console.log("Attempting to sign payload [" + payload + "] with ecdsaKey = [" + JSON.stringify(this.ecdsaKey.privateKey) + "]")
      return await crypto.subtle.sign( {  name: "ECDSA", hash: { name: "SHA-256"}  }, this.ecdsaKey.privateKey, payload )
      .then( async function(signature) {
        var requestOptions = { url: that.provider + payload, headers: hdrs, method: 'HEAD' }
        requestOptions.headers.Sig = Buffer.from(signature).toString('hex')
        requestOptions.headers.Requester = that.ecdsaPublicKeyData
        requestOptions.headers.Msg = payload;
        if (typeof callback === 'function') {
          return request.head(requestOptions, function(err, result, body) {
            return callback(err, body, verifyHeaders(path, hdrs, result.headers));
          });
        } else {
          requestOptions.resolveWithFullResponse = true
          //requestOptions.json = true;
          return rp.head(requestOptions)
            .then( (response) => {
              console.log("Headers inside .head: "  + JSON.stringify(response))
              return response.headers;
            })
            .catch( rperrors.RequestError, function(reason) {
              console.log("ERROR: [head] Request Error - " + reason + " : " + rperrors.RequestError)
              //throw new Error("")
            })
            .catch( (err) => {
              console.log("ERROR: request.head => " + err)
              return
            })
        }
      })
      .catch(function(err) {
        console.error(err);
      });
    }

    parseWolkUrl(wolkurl) {
      let urlParts = wolkurl.split("/");
      let wolkProtocolString = "wolk:"
      var protocol = ""
      var owner = ""
      var collection = ""
      var key = ""
      var wolkUrlObject = {
        protocol: "", owner: "", collection: "", path: "", urltype: "", query: ""
      };
      /* wolk://owner/collection/key */

      if (urlParts.length > 0) {
        wolkUrlObject.protocol = urlParts[0];
        if (wolkUrlObject.protocol != wolkProtocolString) {
          throw new Error("Invalid Wolk URL: " + url)
        }
      }

      if (wolkUrlObject.protocol != "wolk:") {
        console.log('ERROR: calling parseWolkUrl on a non-Wolk URL: ' + wolkurl)
      }

      if (urlParts.length > 2) {
        if (urlParts[2].length > 0) {
          wolkUrlObject.owner = urlParts[2];
        }
      }

      if (urlParts.length > 3) {
        wolkUrlObject.collection = urlParts[3];
      }

      if (urlParts.length > 4) {
        wolkUrlObject.path = wolkurl.substring(wolkUrlObject.protocol.length + wolkUrlObject.owner.length + wolkUrlObject.collection.length + 4);
        wolkUrlObject.urltype = "key"
      }
      //TODO: Query String support -- wolkUrlObject.query = url.query
      return wolkUrlObject;
    }

    buf2hex(buffer) { // buffer is an ArrayBuffer
      return Array.prototype.map.call(new Uint8Array(buffer), x => ('00' + x.toString(16)).slice(-2)).join('');
    }

    //BROWSER version
    async wrequest(method, path, body, hdrs, callback) {
      let that = this;
      if( typeof callback === 'function' ) {
        return this.wreq(method, path, body, hdrs, callback);
      } else {
        return new Promise((resolve, reject) => {
          //console.log("[browser:wrequest] DOWN PROMISE ROUTE")
          return that.wreq(method, path, body, hdrs, function(err, result, proof, verified) {
            if (err && err != 'ERROR: requested proof verification failed') {
              reject(err);
            } else {
              resolve([result, proof, verified]);
            }
          });
        });
      }
    }

    async wreq(method, path, body, hdrs, callback) {
      let that = this
      let payloadBytes = method + "/" + path;
      if ((method == "POST") || (method == "PATCH")) {
        if (typeof body === "object") {
          console.log("[browser:wreq] body was an object so JSON.stringify")
          // console.log(body)          
        }
        if ( body != null ) {
          if (typeof body === "object") {
            payloadBytes += JSON.stringify(body); 
          } else {
            payloadBytes += body;
          }
        }
      }
      let payload = Buffer.from(payloadBytes).toString('utf8')
      var ProofValue = 0

      var stringifyBody = true;
      if (typeof body === "object") {
      } else {
        stringifyBody = false
      }

      // console.log(stringifyBody);
      // console.log(body);


      var requestOptions = {
        method: method,
        url: this.provider + "/" + path,
        headers: {
          "Requester": this.ecdsaPublicKeyData,
          "Msg": payloadBytes
        },
        encoding: null,
        body: body,
        json: stringifyBody,
        resolveWithFullResponse: true
      }
      for (var hdr in hdrs) {
        if (hdr == "Proof" && hdrs["Proof"] == 1) {
          ProofValue = 1;
        }
        requestOptions.headers[hdr] = hdrs[hdr];
      }
      var proof = ''
      var proofType = ''
      var signature

      if( this.ecdsaKey != undefined ) {
        signature = await crypto.subtle.sign( {  name: "ECDSA", hash: { name: "SHA-256"}  }, this.ecdsaKey.privateKey, payload )
                    .catch(function(err) { console.error("ERR: crypto.subtle.sign " + err); })
        requestOptions.headers.Sig = Buffer.from(signature).toString('hex')
      }
      console.log("[browser:wreq] reqopts: " + JSON.stringify(requestOptions))
      return rp(requestOptions)
      .then((response) => {
        if (response.statusCode == "200") {
          //TODO: ERROR HANDLING?
          if (response.headers['proof'] != undefined && response.headers['proof-type'] != undefined) {
            proof = response.headers['proof']
            proofType = response.headers['proof-type']
          }
          //console.log("[browser:wreq] response.body = typeof: " + typeof response.body)
          //console.log(response.body)
          return response.body
        } else {
          console.log("ERROR: StatusCode ", response.statusCode, JSON.stringify(response.status), JSON.stringify(response.statusText))
          // TODO: use response.status
          return callback(new Error(response.body), null);
        }
      })
      .then(function(body) {
        if ( requestOptions.headers.Proof != 1 ) {
          //console.log("[browser:wreq] proof not requested for request")
          //console.log('[browser:wreq] data is: ' + Buffer.from(body))
          //console.log('[browser:wreq] typeof body is: ' + typeof body)
          return callback(null, body);
        } else {
          console.log("Checking Proof")
          if ( proof.length > 0 ) {
            return verifyHeaders(path, hdrs, proof, proofType, body)
            .then( (verified) => {
              if (verified) {
                return callback(null, body, proof, true);
              } else {
                return callback("ERROR: requested proof verification failed", body, proof, false);
              }
            })
            .catch( (err) => { return callback(new Error("unverified proof due to error : " + err), null); })
          } else {
            //console.log(body)
            //console.log("ERROR: proof requested but not returned by wolk provider: ", that.provider)
            return callback(new Error("proof requested but not returned by wolk provider: [" + that.provider + "]"), null);
          }
        }
      })
      .catch(function(err){
        //console.log("ERR: error sending requestOption via rp => " + err + " reqopt: " + JSON.stringify(requestOptions))
        return callback(new Error("Request Error: " + err), null)
      });
    }

    sleep(ms) {    return new Promise(resolve => setTimeout(resolve, ms));    }
    async waitForTx(txhash) {
      console.log("[wolk/browser:waitForTx]")
      // console.log(callback);
      let done = false;
      let calls = 0;
      while (! done) {
        this.getTransaction(txhash).then(([tx]) => {
          // console.log("bbbb")
          console.log(tx);
          try {
            if (tx.BlockNumber) {
              if (! done) {
                done = true;
                return Promise.resolve(tx);    
              } else {
                // console.log("repeat");
              }
            }
          } catch (e) {

          }          
        })
        await this.sleep(500);
        if ( calls++ > 50 ) {
            done = true;
            return Promise.reject(new Error('timeout, tx not included'));
        }
      }
    }
}

function toHexString(byteArray) {
  return Array.prototype.map.call(byteArray, function(byte) {
    return ('0' + (byte & 0xFF).toString(16)).slice(-2);
  }).join('');
}

function toByteArray(hexString) {
  var result = [];
  while (hexString.length >= 2) {
    result.push(parseInt(hexString.substring(0, 2), 16));
    hexString = hexString.substring(2, hexString.length);
  }
  return result;
}

function makeSetKeySignBytes(size, ownerAddress, collection, key) {
  return size + ownerAddress + collection + key;
}

function isValidTx(tx) {
  if (tx.transactionType == TransactionSetName) {
    if (tx.key.length == 0) {
      console.log("[index.js] isValidTx - Invalid Name")
      return false
    }
  }
  if (tx.transactionType == TransactionSetKey) {
    if (tx.collection.length == 0 || tx.collection.length > 128) {
      console.log("[index.js] isValidTx - Invalid Collection Name")
      return false
    }

    if (tx.key.length == 0 || tx.key.length > 128) {
      console.log("[index.js] isValidTx - Invalid Key Length")
      return false
    }
  }
  return true;
}

var joinPath = (...args) => {
  return args.map((part, i) => {
    if (i === 0) {
      return part.trim().replace(/[\/]*$/g, '')
    } else {
      return part.trim().replace(/(^[\/]*|[\/]*$)/g, '')
    }
  }).filter(x => x.length).join('/')
}

function padSize(size) {
  var padding = "";
  var i = 0;
  for (i = 0; i < 16 - size.length; i++) {
    padding += "0";
  }
  return padding + size;
}

function isWolkUrl(url) {
  let protocol = url.split("/")[0];
  if (protocol == "wolk:") {
    return true;
  } else if (protocol == "http:") {
    return true;
  } else {
    throw new Error("Invalid URL protocol: " + url);
  }
}

function computeDefaultHashes(smtTreeDepth) { // (defaultHashes [][]byte) {
  var empty = new Uint8Array(0)
  var defaultHashes = new Array();
  defaultHashes[0] = Computehash(toHexString(empty));
  var level = 0;
  for (level = 1; level < smtTreeDepth; level++) {
    var combinedHashes = Uint8Array.of(...defaultHashes[level - 1], ...defaultHashes[level - 1])
    defaultHashes[level] = Computehash(Buffer.from(combinedHashes));
    // console.log("Calculated Hash [" + level + "]: " + defaultHashes[level].toString('hex'));
  }
  return defaultHashes
}

function hexStringToByte(str) {
  if (!str) {
    return new Uint8Array();
  }
  if ((str[0] == "0") && (str[1] == "x")) {
    str = str.substr(2, str.length - 2);
  }
  var a = [];
  for (var i = 0, len = str.length; i < len; i += 2) {
    a.push(parseInt(str.substr(i, 2), 16));
  }
  return new Uint8Array(a);
}

function byteToHexString(uint8arr) {
  if (!uint8arr) {
    return '';
  }
  var hexStr = '';
  for (var i = 0; i < uint8arr.length; i++) {
    var hex = (uint8arr[i] & 0xff).toString(16);
    hex = (hex.length === 1) ? '0' + hex : hex;
    hexStr += hex;
  }
  return hexStr.toUpperCase();
}

function getBit(k, i) {
  let bitn = i % 8;
  let byten = 20 - 1 - (i >> 3);
  let res = ((1 << bitn) & k[byten]) > 0;
  return res;
}

function keyToAddress(k) {
  return new Uint8Array(Buffer.from(Computehash(k))).subarray(0, 20);
}

function equalBytes(buf1, buf2) {
  if (buf1.byteLength != buf2.byteLength) return false;
  var dv1 = new Int8Array(buf1);
  var dv2 = new Int8Array(buf2);
  for (var i = 0; i != buf1.byteLength; i++) {
    if (dv1[i] != dv2[i]) return false;
  }
  return true;
}

function deserializeProof(sp) {
  let proofhashes = [];
  for (var i = 0; i < sp.Proof.length; i++) {
    proofhashes.push(hexStringToByte(sp.Proof[i]));
  }
  let p = {
    "SMTTreeDepth": 160,
    "Key": hexStringToByte(sp.Key),
    "Proof": proofhashes,
    "ProofBits": hexStringToByte(sp.ProofBits)
  }
  return (p);
}

async function collectionHash(owner, collection) {
  let x = await Computehash(Uint8Array.of(...unicodeStringToTypedArray(owner), ...unicodeStringToTypedArray(collection)));
  return new Uint8Array(x.slice(0, 20))
}

function unicodeStringToTypedArray(s) {
  var escstr = encodeURIComponent(s);
  var binstr = escstr.replace(/%([0-9A-F]{2})/g, function(match, p1) {
    return String.fromCharCode('0x' + p1);
  });
  var ua = new Uint8Array(binstr.length);
  Array.prototype.forEach.call(binstr, function(ch, i) {
    ua[i] = ch.charCodeAt(0);
  });
  return ua;
}

function Computehash(b) {
  return(Buffer.from(toByteArray(sha256(Buffer.from(b)))));
}

// p is a proof structure in raw form like this:
// {"SMTTreeDepth":160,
//   "Key":"c0d5f4f0a1b057d8a1177410e24a857750ac227e",
//    "Proof":["45dab29590cf7c4a54318a33de54934c833cc1594932b98e6566d1c3074a6888","67a193752201548e95f725bd5ab4dfc3cee1d7e84feeebd4bc7bfb826db51ec6"],
//    "ProofBits":"c000000000000000000000000000000000000000"}
// key is 20byte hash, root is 32byte merkle root, defaultHashes is array of 32 byte hashes
function checkSMTProof(p, key, val, root, defaultHashes, verbose) {
  let cur = val;
  let ProofBits = p.ProofBits;
  let pi = 0; // indexes p.Proof
  var combinedHashes;
  for (var i = 0; i < p.SMTTreeDepth; i++) {
    if (getBit(ProofBits, i)) {
      var pb = p.Proof[pi];
      if (getBit(key, i)) {
        //  console.log(" pb1 key 1:", i);
        combinedHashes = Uint8Array.of(...pb, ...cur);
      } else { // i-th bit is "0", so hash with H([]) on the right
        //  console.log(" pb1 key 0:", i);
        combinedHashes = Uint8Array.of(...cur, ...pb);
      }
      cur = Computehash(Buffer.from(combinedHashes));
      pi++;
    } else {
      if (getBit(key, i)) { // i-th bit is "1", so hash with H([]) on the left
        //  console.log(" pb0 key 1:", i);
        combinedHashes = Uint8Array.of(...defaultHashes[i], ...cur);
      } else {
        //  console.log(" pb0 key 0:", i);
        combinedHashes = Uint8Array.of(...cur, ...defaultHashes[i]);
      }
      cur = Computehash(Buffer.from(combinedHashes));
    }
  }
  let merkleRootComputed = byteToHexString(cur);
  let merkleRootExpected = byteToHexString(root);
  if (merkleRootComputed == merkleRootExpected) {
    if (verbose > 1) console.log("SUCCESS", "merkleRootComputed", merkleRootComputed);
    // TODO: check that this merkleRoot is actually IN the block with GetBlock
    return true;
  }
  console.log("checkSMTProof FAIL-merkleRoot MISMATCH", merkleRootComputed, "merkleRootExpected", merkleRootExpected);
  return false;
}

function verifyHeaders(path, requestHeaders, proof, proofType, body) {
  if (requestHeaders.Proof != undefined) {
    switch (proofType) {
      case "SMT":
        let smtProof = JSON.parse(proof)
        return verifySMTProof(smtProof, 0);
      case "NoSQLScan":
      case "NoSQL":
        let nosqlProof = JSON.parse(proof)
        return verifyNoSQLProof(proofType, nosqlProof, path, 0, body, 0);
    }
  } else {
    // console.log("requested proof, but no proof in response");
    return true;
  }
  return false;
}

async function verifySMTProof(proofObject, verbose) {
  let proof = deserializeProof(proofObject.Proof);
  return await checkSMTProof(proof,
    hexStringToByte(proofObject.KeyHash),
    hexStringToByte(proofObject.TxHash),
    hexStringToByte(proofObject.MerkleRoot),
    verbose);
}

function verifyContentHash(body) {
  //TODO
  return true;
}

async function verifyNoSQLProof(proofType, proofObject, path, contentLen, body, verbose) {
  let wolkUrlObject = parseWolkPath(path)
  if (proofObject.Collection != wolkUrlObject.collection) {
    console.log("lproofObject")
    console.log(proofObject)
    console.log("wolkUrlObject")
    console.log(wolkUrlObject)
    console.log("Collection Mismatch", proofObject.Collection, wolkUrlObject.collection)
    return false;
  }
  let defaultHashes = await computeDefaultHashes(160);

  if (proofType == "NoSQLScan") {
    if ( proofObject.ScanProofs != undefined ) {
      console.log("NoSQLScan with ScanProofs")
      for (var i = 0; i < proofObject.ScanProofs.length; i++) {
        let p = proofObject.ScanProofs[i];
        let prf = deserializeProof(p.KeyProof);
        if (proofObject.Collection.length == 0) {
          // TODO: add await
          let smtCheck = await checkSMTProof(prf, prf.Key, hexStringToByte(p.TxHash), hexStringToByte(proofObject.SystemMerkleRoot), defaultHashes, verbose)
          if (!smtCheck) {
            return false;
          }
        } else {
          let smtCheck = checkSMTProof(prf, prf.Key, hexStringToByte(p.TxHash), hexStringToByte(proofObject.CollectionMerkleRoot), defaultHashes, verbose)
          if (!smtCheck) {
            return false;
          }
        }
      }
    } else {
      if (proofObject.Collection.length == 0) { //listCollections
        if ( ! ( proofObject.SystemChunkHash === "0x44a00e7ae0499c6e377b95709f3851843976ea6c2a86eff46adfe15608d22005") ) {
          return false;
        }
        let prf = deserializeProof(proofObject.SystemProof)
        let smtCheck = checkSMTProof(prf, prf.Key, hexStringToByte(proofObject.SystemChunkHash), hexStringToByte(proofObject.KeyMerkleRoot), defaultHashes, verbose)
        if (!smtCheck) {
          return false;
        }
        console.log("NoSQLScan SUCCESS", "ch", proofObject.SystemChunkHash, "mr", proofObject.KeyMerkleRoot)
      } else { //scanCollection
        if ( ! ( proofObject.CollectionChunkHash === "0x44a00e7ae0499c6e377b95709f3851843976ea6c2a86eff46adfe15608d22005" ) ) {
          return false;
        }
      }
    }
    console.log("NoSQLScan Proof Verified")
    return true;
  }

  if (proofObject.Key != wolkUrlObject.path) {
    console.log(proofObject)
    console.log("Key Mismatch: [", proofObject.Key, "] key: [", wolkUrlObject.path, "]")
    return false;
  }

  // Verify Proof Process: at proof.Key position (which is keyToAddress(Key) where key is supplied by user)
  //   there is proof.TxHash, which hashes up to proof.CollectionMerkleRoot
  //console.log("proofObject.KeyProof: " + JSON.stringify(proofObject) )
  let keyProof = deserializeProof(proofObject.KeyProof);
  let keyHash = await keyToAddress(wolkUrlObject.path);
  if (!equalBytes(keyHash, keyProof.Key)) {
    console.log("Key Mismatch", keyProof.Key, "keyHash", keyHash, "path", wolkUrlObject.path);
    return false;
  }

  // position: keyhash, txhash = value hashes up to merkleroot
  let proof_txHash = hexStringToByte(proofObject.TxHash);
  let proof_collectionMerkleRoot = hexStringToByte(proofObject.CollectionMerkleRoot);
  let keyProofCheck = await checkSMTProof(keyProof, keyHash, proof_txHash, proof_collectionMerkleRoot, defaultHashes, verbose)
  if (!keyProofCheck) {
    console.log("KeyProof Verification FAILED");
    return false;
  }
  console.log("KeyProofCheck: ", keyProofCheck);
  if (proofObject.Tx == undefined) {
    console.log("Tx not supplied")
    return false;
  }

  // check that txhash matches provided tx ... TODO: check that pr.TxHash matches the Tx returned ***in RLP encoded form***
  let txhash = hexStringToByte(proofObject.Tx.txhash);
  let owner = wolkUrlObject.owner;
  let txsigner = hexStringToByte(proofObject.Tx.signer);
  /*
    if (!equalBytes(txhash, proof_txHash)) {
      console.log("Tx mismatch", txhash, proof_txHash);
      return false;
    }
  let txAmount = proofObject.Tx.amount;
  let contentLength = contentLen;
  if (contentLength != txAmount) {
    console.log("Header:Content-Length Does not Match Tx.Amount", "Header:Content-Length", contentLength, "Tx.Amount", txAmount);
    return false;
  }
  // TODO: check that TxSigner is valid, matches owner
  if (!equalBytes(txsigner, owner)) {
    // TODO: check against writers list?
    console.log("Signer mismatch");
    return false
  }
  if (!verifyContentHash(proofObject.Tx.GenericHash, body)) {
    console.log("ContentHash Mismatch");
    return false;
  }
  */

  // Verify Proof Process: at proof.Key position (which is proof.CollectionHash=CollectionHash(owner, collection))
  //   there is proof.CollectionChunkHash, which hashes up to proof.KeyMerkleRoot
  let collectionProof = deserializeProof(proofObject.CollectionProof);
  let collHash = await collectionHash(owner, wolkUrlObject.collection)
  if (!equalBytes(collectionProof.Key, collHash)) {
    console.log("labelcollectionProof")
    console.log(collectionProof)
    console.log("labelcollHash")
    console.log(collHash)
    console.log("Collection Mismatch", collectionProof.Key.toString(), "collHash", collHash.toString());
    return false;
  }

  let proof_collectionChunkHash = hexStringToByte(proofObject.CollectionChunkHash);
  let proof_keyMerkleRoot = hexStringToByte(proofObject.KeyMerkleRoot);
  let collectionCheck = await checkSMTProof(collectionProof, collHash, proof_collectionChunkHash, proof_keyMerkleRoot, defaultHashes, verbose)
  if ( ! collectionCheck ) {
    console.log("CollectionProof Verification FAILED", collectionProof);
    return false;
  }
  console.log("CollectionProof Verified");

  // Verify proof process: at proof.Key position (which is proof.SystemHash=CollectionHash(owner, SystemCollection))
  //   there is the value proof.SystemChunkHash, which hashes up to proof.KeyMerkleRoot
  let systemProof = deserializeProof(proofObject.SystemProof)
  let systemHash = await keyToAddress(owner)
  if (!equalBytes(systemProof.Key, systemHash)) {
    console.log("System Mismatch")
    return false;
  }
  let proof_systemChunkHash = hexStringToByte(proofObject.SystemChunkHash);
  //let proof_keyMerkleRoot = hexStringToByte(proofObject.KeyMerkleRoot);
  let ownerCheck = await checkSMTProof(systemProof, systemHash, proof_systemChunkHash, proof_keyMerkleRoot, defaultHashes, verbose)
  if ( ! ownerCheck ) {
    console.log("SystemProof FAILED")
    return false;
  }
  console.log("SystemProof Verified");
  return true;
}

function parseWolkUrl(wolkurl) {
  let urlParts = wolkurl.split("/");
  let wolkProtocolString = "wolk:"
  var protocol = ""
  var owner = ""
  var collection = ""
  var key = ""
  var wolkUrlObject = {
    protocol: "",
    owner: "",
    collection: "",
    path: "",
    urltype: "",
    query: ""
  };
  /* wolk://owner/collection/key */

  if (urlParts.length > 0) {
    wolkUrlObject.protocol = urlParts[0];
    if (wolkUrlObject.protocol != wolkProtocolString) {
      throw new Error("Invalid Wolk URL: " + url)
    }
  }

  if (wolkUrlObject.protocol != "wolk:") {
    console.log('ERROR: calling parseWolkUrl on a non-Wolk URL: ' + wolkurl)
  }

  if (urlParts.length > 2) {
    if (urlParts[2].length > 0) {
      wolkUrlObject.owner = urlParts[2];
    } else {
      //throw new Error("Invalid Wolk URL: " + wolkurl + "owner length must be > 0")
      //though we could just show wolk://
    }
  }

  if (urlParts.length > 3) {
    wolkUrlObject.collection = urlParts[3];
  }

  if (urlParts.length > 4) {
    wolkUrlObject.path = wolkurl.substring(wolkUrlObject.protocol.length + wolkUrlObject.owner.length + wolkUrlObject.collection.length + 4);
    wolkUrlObject.urltype = "key"
  }
  //TODO: Query String support -- wolkUrlObject.query = url.query
  return wolkUrlObject;
}
function parseWolkPath(wolkPath) {
  let urlParts = wolkPath.split("/");
  var owner = ""
  var collection = ""
  var key = ""
  var wolkUrlObject = {
    owner: "",
    collection: "",
    path: "",
  };
  /* i.e. owner/collection/key -- NOTE: Different from parseWolkUrl */

  if (urlParts.length > 0) {
      wolkUrlObject.owner = urlParts.shift();
  }
  if (urlParts.length > 0) {
      wolkUrlObject.collection = urlParts.shift();
  }
  if (urlParts.length > 0) {
      wolkUrlObject.path = urlParts.join("/")
  }
  return wolkUrlObject;
}


if (typeof window !== 'undefined' && typeof window.WOLK === 'undefined') {
    window.WOLK = WOLK;
}
module.exports = WOLK;
