'use strict';
const request = require('request');
const rp = require('request-promise');
const rperrors = require('request-promise/errors');
const path = require('path');
const crypto = require('@trust/webcrypto')
const fs = require('fs');
const sha256 = require('sha256');
const WolkError = require('./wolkerror.js')

const defaultAccountFile = "default"
const ecdsaPublicKeyFile = "public.key"
const ecdsaPrivateKeyFile = "private.key"
const rsaPublicKeyFile = "id_rsa.pub"
const rsaPrivateKeyFile = "id_rsa"
const friendsKeyFile = "friends.key"
const personalKeyFile = "personal.key"

const ProtocolName = "wolk"
const PayloadChunk = "chunk"

const SystemCollection = "buckets";
const SMTTreeDepth = 160;

/* TransactionType Constants */
const TransactionTransfer = 1
const TransactionRegisterNode = 2
const TransactionUpdateNode = 3
const TransactionBandwidthCheck = 4
const TransactionSetQuota = 5
const TransactionSetKey = 6
const TransactionDeleteKey = 7
const TransactionSetName = 8
const TransactionMutate = 9

/* BucketType Constants */
const BucketCloudstore = 0
const BucketNoSQL = 1
const BucketSQL = 2
const DefaultMaxBucketSize = 1000000

/* SQLREQUEST OPTION Constants */
const RT_PUT = 'Put'
const RT_GET = "Get"
const RT_DELETE = "Delete"
const RT_QUERY = "Query"
const RT_SCAN = "Scan"
const DEFAULT_DEVELOPER_TRUST_LEVEL = 1 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time
const DEFAULT_USER_TRUST_LEVEL = 1 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time

class WOLK {
  constructor() {
    this.provider = "https://cloud.wolk.com";
    this.keyDir = "~/.wolk/keys"
    this.developerTrustLevel = DEFAULT_DEVELOPER_TRUST_LEVEL
    this.userTrustLevel = DEFAULT_USER_TRUST_LEVEL
    this.setWolkTrustLevel()
  }

  async init() {
    return await this.loadDefaultAccount()
      .then( (account) => {
        return account
      })
      .catch( (err) => {
        console.log("WOLK: loadAccount error: " + err)
        throw err;
      })
  }

  setKeyDir(_keyDir) {
      this.keyDir = _keyDir
  }

  setProvider( provider ) {
    this.provider = provider
  }

  setDeveloperTrustLevel( level ) {
    this.developerTrustLevel = level
  }

  setUserTrustLevel( level ) {
    this.userTrustLevel = level
  }

  setWolkTrustLevel() {
    if( this.userTrustLevel > this.developerTrustLevel ) {
      this.wolkTrustLevel  = this.userTrustLevel
    } else {
      this.wolkTrustLevel  = this.developerTrustLevel
    }
  }

  // account operations
  loadDefaultAccount() {
    let fn = path.join(this.keyDir, defaultAccountFile)
    if (fs.existsSync(fn)) {
      let defaultAccount = fs.readFileSync(fn, 'utf8');
      // console.log("Default account: ", defaultAccount)
      return this.loadAccount(defaultAccount)
              .then( () => {
                //console.log("[fs:loadDefaultAccount] loadAccount - loaded: " + defaultAccount)
                return defaultAccount
              })
              .catch( (err) => {
                throw new WolkError(0, "Unable to load defaultAccount [" + defaultAccount + "]")
              })
    } else {
      throw new WolkError(0, "No Default Account found: please create one")
    }
    return new Promise(function(resolve, reject) {
        resolve("");
    });
  }

  async setDefaultAccount( name ) {
    name = name.toLowerCase();
    let fn = path.join(this.keyDir, defaultAccountFile)
    return await p_writefile(fn, name, 'utf8')
      .then( () => {
        //console.log('SUCC: Setting default account')
        this.defaultAccount = name
        //console.log("SetDefault")
        return new Promise( function(resolve, reject) {
            resolve(name);
          });
      })
  }

  async createAccount(name, callback) {
    name = name.toLowerCase();
    // we assume account not on blockchain
    let accountDir = path.join(this.keyDir, name)
    if (fs.existsSync(accountDir)) {
      console.log("Account Already Exists Setting as DEFAULT:", accountDir)
      //console.log("Provider: " + this.provider)
      return await this.setDefaultAccount( name )
      .then( () => {
        //console.log("loading default")
        return this.loadDefaultAccount()
        .then(() => {
          return this.loadAccount(name)
          .then( () => { return "accountexistslocally"; })
          .catch( (err) => { console.error("[fs:createAccount] Error - loadAccount: " + name + " | " + err); });
        });
      })
    } else {
      console.log("Creating necessary key files locally")
      fs.mkdirSync(accountDir, { recursive: true }, (err) => {
        if (err) throw err;
      });

      // generate RSA + ECDSA Keys
      let rsaKey = await crypto.subtle.generateKey({
        name: "RSASSA-PKCS1-v1_5",
        modulusLength: 2048,
        publicExponent: new Uint8Array([0x01, 0x00, 0x01]),
        hash: { name: "SHA-256" }
      }, true, ["sign", "verify"]);
      let ecdsaKey = await crypto.subtle.generateKey({
        name: "ECDSA",
        namedCurve: "P-256"
      }, true, ["sign", "verify"]);

      let friendsKey = await crypto.subtle.generateKey(
        {
            name: "AES-CTR",
            length: 256, //can be  128, 192, or 256
        },
        true,
        ["encrypt", "decrypt"] //can "encrypt", "decrypt", "wrapKey", or "unwrapKey"
      )

      let personalKey = await crypto.subtle.generateKey(
        {
            name: "AES-CTR",
            length: 256, //can be  128, 192, or 256
        },
        true,
        ["encrypt", "decrypt"] //can "encrypt", "decrypt", "wrapKey", or "unwrapKey"
      )
      this.rsaKey = rsaKey.privateKey
      this.ecdsaKey = ecdsaKey.privateKey
      this.friendsKey = friendsKey
      this.personalKey = personalKey

      // export RSA + ECDSA Keys
      let rsaPublicKey = await crypto.subtle.exportKey("jwk", rsaKey.publicKey)
      let rsaPrivateKey = await crypto.subtle.exportKey("jwk", rsaKey.privateKey)
      let ecdsaPublicKey = await crypto.subtle.exportKey("jwk", ecdsaKey.publicKey)
      let ecdsaPrivateKey = await crypto.subtle.exportKey("jwk", ecdsaKey.privateKey)
      let friendsExportKey = await crypto.subtle.exportKey("jwk", friendsKey)
      let personalExportKey = await crypto.subtle.exportKey("jwk", personalKey)

      this.ecdsaPublicKey = JSON.stringify(ecdsaPublicKey);
      this.ecdsaPrivateKey = JSON.stringify(ecdsaPrivateKey);
      this.rsaPrivateKey = JSON.stringify(rsaPrivateKey);
      this.rsaPublicKey = JSON.stringify(rsaPublicKey);
      this.friendsExportKey = JSON.stringify(friendsExportKey);
      this.personalExportKey = JSON.stringify(personalExportKey);

      // save files
      return p_writefile(path.join(accountDir, rsaPrivateKeyFile), this.rsaPrivateKey, 'utf8')
      .then( () => { return p_writefile(path.join(accountDir, rsaPublicKeyFile), this.rsaPublicKey, 'utf8') })
      .then( () => { return p_writefile(path.join(accountDir, ecdsaPublicKeyFile), this.ecdsaPublicKey, 'utf8') })
      .then( () => { return p_writefile(path.join(accountDir, ecdsaPrivateKeyFile), this.ecdsaPrivateKey, 'utf8') })
      .then( () => { return p_writefile(path.join(accountDir, friendsKeyFile), this.friendsExportKey, 'utf8') })
      .then( () => { return p_writefile(path.join(accountDir, personalKeyFile), this.personalExportKey, 'utf8') })
      .then( () => {
        //console.log("Saved keys")
        if (this.defaultAccount == undefined) {
          this.defaultAccount = name
          let fn = path.join(this.keyDir, defaultAccountFile)
          return p_writefile(fn, name, 'utf8')
        }
        return new Promise((resolve, reject) => { resolve("") });
      })
      .then( async () => {
        return await this.loadDefaultAccount(name).then( async () => {
                 let setNameProm = await this.setName(name, JSON.stringify(rsaPublicKey), callback)
                 return setNameProm
                });
      })
      .catch( (err) => { throw new WolkError(0, "ERROR: createAccount " + err) });
    }
  }

  loadAccount(name) {
    name = name.toLowerCase();
    let accountDir = path.join(this.keyDir, name)
    if (!fs.existsSync(accountDir)) {
      console.log("no such directory:", accountDir)
    }
    this.ecdsaPublicKey = fs.readFileSync(path.join(accountDir, ecdsaPublicKeyFile), 'utf8');
    this.ecdsaPrivateKey = fs.readFileSync(path.join(accountDir, ecdsaPrivateKeyFile), 'utf8');
    this.rsaPrivateKey = fs.readFileSync(path.join(accountDir, rsaPrivateKeyFile), 'utf8');
    this.rsaPublicKey = fs.readFileSync(path.join(accountDir, rsaPublicKeyFile), 'utf8');
    this.friendsExportKey = fs.readFileSync(path.join(accountDir, friendsKeyFile), 'utf8');
    this.personalExportKey = fs.readFileSync(path.join(accountDir, personalKeyFile), 'utf8');

    return crypto.subtle.importKey("jwk", JSON.parse(this.ecdsaPrivateKey), {
        name: "ECDSA",
        namedCurve: "P-256"
      }, false, ['sign'])
      .then((result, err) => {
        this.ecdsaKey = {}
        this.ecdsaKey.privateKey = result;
        return
      })
      .catch((error) => {
        //console.log(" Error Getting ECDSA Private key: " + error);
      });
  }

  setName(name, rsaPublicKey, options, callback) {
    name = name.toLowerCase();
    return this.post(name, JSON.stringify({
      "name": name,
      "rsaPublicKey": Buffer.from(rsaPublicKey).toString('base64')
    }), options, callback);
  }

  determineProofValue() {
    let randProof = Math.random();
    if ( randProof <= this.wolkTrustLevel ) {
      return 1
    } else {
      return 0
    }
  }

  // Blockchain operations
  getBlock(blockNumber, callback) {
    return this.get(joinPath('wolk', 'block', blockNumber.toString(10)), { "Proof": this.determineProofValue() }, callback)
            .catch( (err) => { throw err; });
  }
  getNode(nodeNumber, callback) {
    return this.get(joinPath('wolk', 'node', nodeNumber.toString(10)), { "Proof": this.determineProofValue() }, callback)
            .catch( (err) => { throw err; });
  }

  getPeers(nodeNumber, callback) {
    return this.get(joinPath('wolk', 'info'), { "Proof": this.determineProofValue() }, callback)
      .then( (info) => {
        info = info.toString();
        if( typeof info == 'string' ) {
          info = JSON.parse(info);
        }
        return info.peerlist
      })
      .catch( (err) => {
        console.error("error: ", err);
        throw err;
      });
  }

  getNumPeers(nodeNumber, callback) {
    return this.get(joinPath('wolk', 'info'), { "Proof": this.determineProofValue() }, callback)
      .then( (info) => {
        info = info.toString();
        if( typeof info == 'string' ) {
          info = JSON.parse(info)
        }
        return info.peers
      })
      .catch( (err) => { throw err; });
  }
  getPendingTxCount(nodeNumber, callback) {
    return this.get(joinPath('wolk', 'info'), { "Proof": this.determineProofValue() }, callback)
      .then( (info) => {
        info = info.toString();
        if( typeof info == 'string' ) {
          info = JSON.parse(info)
        }
        return info.pendingTxCount
      })
      .catch( (err) => { throw err; });
  }
  updateNode(nodeNumber, nodeInfo, callback) {
    return this.patch(joinPath('wolk', 'node', nodeNumber.toString(10)), nodeInfo, {}, callback);
  }
  getName(name, callback) {
    name = name.toLowerCase();
    return this.get(joinPath('wolk', 'name', name), {
      "Proof": this.determineProofValue()
    }, callback)
    .catch( (err) => { throw err; });
  }
  getAccount(name, callback) {
    name = name.toLowerCase();
    return this.get(joinPath('wolk', 'account', name), {
      "Proof": this.determineProofValue()
    }, callback)
    .catch( (err) => { throw err; });
  }
  getLatestBlockNumber(callback) {
    return this.get(joinPath('wolk', 'block', 'latest'), { "Proof": this.determineProofValue() }, callback)
      .catch( (err) => { throw err; });
  }
  getBalance(name_address, callback) {
    name_address = name_address.toLowerCase();
    return this.getAccount(name_address)
            .then( accObject => {
              var account = JSON.parse(accObject)
              return account.balance
            })
            .catch( (err) => { throw err; });
  }
  getTransaction(txhash, callback) {
    return this.get(joinPath('wolk', 'tx', txhash), { "Proof": this.determineProofValue() }, callback)
      .catch( (err) => { throw err; });
  }
  transfer(recipient, amount, callback) {
    return this.post(joinPath('wolk', 'transfer'), {
      "recipient": recipient,
      "amount": amount
    }, {}, callback);
  }
  // NoSQL
  createBucket(owner, collection, options, callback) {
    var txBucket = NewTxBucket(collection, BucketCloudstore, options)
    return this.post(joinPath(owner, collection), JSON.stringify(txBucket), options, callback);
  }
  createCollection(owner, collection, options, callback) {
    var txBucket = NewTxBucket(collection, BucketNoSQL, options)
    return this.post(joinPath(owner, collection), JSON.stringify(txBucket), options, callback);
  }
  listCollections(owner, options, callback) {
    return this.get(owner, options, callback)
      .catch( (err) => { throw err; });
  }
  listCollection(owner, collection, options, callback) {
    return this.get(joinPath(owner, collection), options, callback)
      .catch( (err) => { throw err; });
  }
  deleteCollection(owner, collection, options, callback) {
    return this.delete(joinPath(owner, collection), options, callback);
  }
  scanCollection(owner, collection, hdrs, options, callback) {
    return this.get(joinPath(owner, collection), options, callback)
      .catch( (err) => { throw err; });
  }
  setKey(owner, collection, key, value, options, callback) {
    return this.put(joinPath(owner, collection, key), value, options, callback);
  }
  getKey(owner, collection, key, hdrs, callback) {
    //console.log("Getting key: owner " + owner + " | collection " + collection + " | key " + key)
    return this.get(joinPath(owner, collection, key), hdrs, callback).catch( (err) => { throw err; });
  }

  getChunk( chunkId, options, callback) {
    return this.get(joinPath(ProtocolName, PayloadChunk, chunkId), { "Proof": this.determineProofValue() }, callback)
            .catch( (err) => { throw err; });
  }

  deleteKey(owner, collection, key, options, callback) {
    return this.delete(joinPath(owner, collection, key), options, callback);
  }

  // SQL
  createDatabase(owner, database, options, callback) {
    return this.post(joinPath(owner, database), JSON.stringify({
      "name": database,
      "bucketType": BucketSQL,
      "requesterPays": 0,
      "encryption": "none"
    }), options, {}, callback);
  }
  listDatabases(owner, callback) {
    return this.get(joinPath(owner), options, callback)
      .catch( (err) => { throw err; });
  }
  deleteDatabase(owner, database, callback) {
    return this.delete(joinPath(owner, collection), {}, callback);
  }
  executeSQL(owner, database, sql, hdrs, callback) {
    return this.post(joinPath(owner, database, sql), sql, hdrs, callback);
  }

  // NOT EXPOSED
  async get(path, hdrs, callback) {
    return this.wrequest("GET", path, null, hdrs, callback);
  }

  async post(path, body, hdrs, callback) {
    return this.wrequest("POST", path, body, hdrs, callback);
  }

  async put(path, body, hdrs, callback) {
    return this.wrequest("PUT", path, body, hdrs, callback);
  }

  async patch(path, body, hdrs, callback) {
    return this.wrequest("PATCH", path, body, hdrs, callback);
  }

  async delete(path, hdrs, callback) {
    return this.wrequest("DELETE", path, null, hdrs, callback);
  }

  sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
  }

  computeHash(b) {
    return(Buffer.from(this.toByteArray(sha256(Buffer.from(b)))));
  }

  toByteArray(hexString) {
    var result = [];
    while (hexString.length >= 2) {
      result.push(parseInt(hexString.substring(0, 2), 16));
      hexString = hexString.substring(2, hexString.length);
    }
    return result;
  }

  async waitForTx(txhash, callback) {
    let done = false;
    let calls = 0;
    while (!done) {
      this.getTransaction(txhash, function(err, body, verified) {
        if (err) {
          //console.log("err:", err);
        } else {
          try {
            let tx = JSON.parse(body)
            if (tx.BlockNumber != undefined) {
              done = true;
              return callback(null, tx);
            }
          } catch (e) {

          }
        }
      })
      await this.sleep(500);
      if (calls++ > 15) {
        done = true;
        return callback(new Error('timeout, tx not included'), null);
      }
    }
  }

  async wrequest(method, path, body, hdrs, callback) {
    let that = this;
    if( typeof callback === 'function' ) {
      return this.wreq(method, path, body, hdrs, callback);
    } else {
      return new Promise((resolve, reject) => {
        return that.wreq(method, path, body, hdrs, function(err, result) {
          if (err) {
            reject(err);
          } else {
            resolve(result);
          }
        });
      });
    }
  }

  async wreq(method, path, body, hdrs, callback) {
    let payloadBytes = method + "/" + path;
    var encodingValue = "utf8"
    if ( (method == "POST") || (method == "PATCH") ) {
      if (typeof body === "object") {
        //console.log("[fs:wreq] body was an object so JSON.stringify")
        //console.log(body)
        //body = JSON.stringify(body);
      }
      if ( body != null ) { payloadBytes += body; }
      encodingValue = null
    }
    let payload = Buffer.from(payloadBytes).toString('utf8')
    let that = this
    var stringifyBody = true;
    if (typeof body === "object") {
    } else {
      stringifyBody = false
    }
    var requestOptions = {
      method: method,
      url: this.provider + "/" + path,
      headers: {
        "Requester": this.ecdsaPublicKey,
        "Msg": payloadBytes
      },
      encoding: null,
      body: body,
      //json: false, //stringifyBody,
      resolveWithFullResponse: true
    }

    for (var hdr in hdrs) {
      requestOptions.headers[hdr] = hdrs[hdr];
    }
    var proof = ''
    var proofType = ''
    var signature

    if ( this.ecdsaKey != undefined ) {
      signature = await crypto.subtle.sign({ name: "ECDSA", hash: { name: "SHA-256" } }, this.ecdsaKey.privateKey, payload)
                  .catch(function(err) { console.error("ERR: crypto.subtle.sign " + err); });
      requestOptions.headers.Sig = buf2hex(signature);
      // console.log('payload')
      // console.log(payload)
      // console.log('raw sig')
      // console.log(signature)
      // console.log('sig buf2hex')
      // console.log(requestOptions.headers.Sig)
    } else {
      //console.warn("WARN: ecdsaKey is undefined");
    }

	  //console.log("RequestOption: " + JSON.stringify(requestOptions));
    return rp(requestOptions)
    .then((response) => {
      //console.log("\nprocessing response: " + JSON.stringify(response))
      if (response.statusCode == "200") {
        //TODO: ERROR HANDLING?
        if (response.headers['proof'] != undefined && response.headers['proof-type'] != undefined) {
          //console.log("Response headers = " + JSON.stringify(response.headers))
          proof = response.headers['proof']
          proofType = response.headers['proof-type']
        }
        return response.body
      } else {
        console.log("ERROR: StatusCode ", response.statusCode, JSON.stringify(response.status), JSON.stringify(response.statusText))
        // TODO: use response.status
        return callback(new WolkError(response.statusCode, response.body), null);
      }
    })
    .then(function(body) {
      //console.log("\nprocessing body: " + JSON.stringify(body))
      if ( requestOptions.headers.Proof != 1 ) {
        if( method == "GET" ) {
          //console.log("proof not requested for request")
        }
        return callback(null, body);
      } else {
        //console.log("Checking Proof")
        if ( proof.length > 0 ) {
          //console.log("\nPROOF: " + JSON.stringify(proof))
          return verifyHeaders(path, hdrs, proof, proofType, body)
          .then( (verified) => {
            if (verified) {
              //console.log("body verified")
              return callback(null, body);
            } else {
              //console.log("body NOTverified")
              let statusCode = 412
              return callback(new WolkError(statusCode, "unverified proof"), body);
            }
          })
          let statusCode = 412
          .catch( (err) => { return callback(new WolkError(statusCode, "unverified proof due to error : " + err), null); })
        } else {
          //console.log(body)
          //console.log("ERROR: proof requested but not returned by wolk provider: ", that.provider)
          let statusCode = 412
          return callback(new WolkError(statusCode, "proof requested but not returned by wolk provider: [" + this.provider + "]"), null);
        }
      }
    })
    .catch(rperrors.StatusCodeError, function(reason) {
      var error = new WolkError(reason.statusCode, reason.error )
      return callback( error, null ) //+ " reqopt: " + JSON.stringify(requestOptions)))
    })
    .catch( (err) => {
      //console.log(err)
      let statusCode = 503
      return callback(new WolkError(statusCode, err), null);
    })
    //.catch( rperrors.RequestError, function(reason) { console.error("ERROR: 1RequestError " + reason.cause); })
  }
}

function NewTxBucket(name, bucketType, options) {
    var txBucket = {
      name: name,
      bucketType: bucketType
    };
    if( options != undefined ) {
      txBucket.requesterPays = 0; //options.requesterPays
      //TODO: txBucket.writers = options.validWriters
      //TODO: txBucket.schemaURL = options.schema
      txBucket.encryption = "none"; //options.encryption
      txBucket.shimURL = options.shimURL
    }
    return txBucket
}

function p_writefile(path, data, encoding) {
  return new Promise(function(resolve, reject) {
      fs.writeFile(path, data, encoding, function(err) {
          if (err) { reject(err); }
          else { resolve(data); }
      });
  });
}

function buf2hex(buffer) { // buffer is an ArrayBuffer
  return Array.prototype.map.call(new Uint8Array(buffer), x => ('00' + x.toString(16)).slice(-2)).join('');
}

function joinPath (...args) {
  return args.map((part, i) => {
    if (i === 0) {
      return part.trim().replace(/[\/]*$/g, '')
    } else {
      return part.trim().replace(/(^[\/]*|[\/]*$)/g, '')
    }
  }).filter(x => x.length).join('/')
}

function Computehash(b) {
  return crypto.subtle.digest({ name: "SHA-256" }, b)
    .catch( (err) => {
      console.error("[fs.js:ComputeHash] error calling crypto.subtle.digest => " + err)
    })
}

function hexStringToByte(str) {
  if (!str) {
    return new Uint8Array();
  }
  if ((str[0] == "0") && (str[1] == "x")) {
    str = str.substr(2, str.length - 2);
  }
  var a = [];
  for (var i = 0, len = str.length; i < len; i += 2) {
    a.push(parseInt(str.substr(i, 2), 16));
  }
  return new Uint8Array(a);
}

function getBit(k, i) {
  let bitn = i % 8;
  let byten = 20 - 1 - (i >> 3);
  let res = ((1 << bitn) & k[byten]) > 0;
  return res;
}

function equalBytes(buf1, buf2) {
  if (buf1.byteLength != buf2.byteLength) return false;
  var dv1 = new Int8Array(buf1);
  var dv2 = new Int8Array(buf2);
  for (var i = 0; i != buf1.byteLength; i++) {
    if (dv1[i] != dv2[i]) return false;
  }
  return true;
}

function deserializeProof(sp) {
  let proofhashes = [];
  for (var i = 0; i < sp.Proof.length; i++) {
    proofhashes.push(hexStringToByte(sp.Proof[i]));
  }
  let p = {
    "SMTTreeDepth": 160,
    "Key": hexStringToByte(sp.Key),
    "Proof": proofhashes,
    "ProofBits": hexStringToByte(sp.ProofBits)
  }
  return (p);
}

async function computeDefaultHashes(smtTreeDepth) {
  var defaultHashes = new Array();
  defaultHashes[0] = new Uint8Array(await Computehash(new Uint8Array(0)));
  for (var level = 1; level < smtTreeDepth; level++) {
    defaultHashes[level] = await Computehash(Uint8Array.of(...new Uint8Array(defaultHashes[level - 1]), ...new Uint8Array(defaultHashes[level - 1])));
  }
  return defaultHashes
}

async function checkSMTProof(p, key, val, root, verbose) {
  var defaultHashes = await computeDefaultHashes(160)
  let cur = val;
  let ProofBits = p.ProofBits;
  let pi = 0; // indexes p.Proof
  var combinedHashes;
  for (var i = 0; i < p.SMTTreeDepth; i++) {
    if (getBit(ProofBits, i)) {
      var pb = p.Proof[pi];
      if (getBit(key, i)) {
        cur = await Computehash(Uint8Array.of(...pb, ...new Uint8Array(cur)));
      } else { // i-th bit is "0", so hash with H([]) on the right
        cur = await Computehash(Uint8Array.of(...new Uint8Array(cur), ...pb));
      }
      pi++;
    } else {
      if (getBit(key, i)) { // i-th bit is "1", so hash with H([]) on the left
        cur = await Computehash(Uint8Array.of(...new Uint8Array(defaultHashes[i]), ...new Uint8Array(cur)));
      } else {
        cur = await Computehash(Uint8Array.of(...new Uint8Array(cur), ...new Uint8Array(defaultHashes[i])));
      }
    }
  }
  let merkleRootComputed = buf2hex(cur);
  let merkleRootExpected = buf2hex(root);
  if (merkleRootComputed == merkleRootExpected) {
    if (verbose > 1) console.log("SUCCESS", "merkleRootComputed", merkleRootComputed);
    // TODO: check that this merkleRoot is actually IN the block with GetBlock
    return true;
  }
  console.log("checkSMTProof FAIL-merkleRoot MISMATCH", merkleRootComputed) //  "merkleRootExpected", merkleRootExpected);
  return false;
}

function verifyHeaders(path, requestHeaders, proof, proofType, body) {
  //console.log("RequestHeaders: " + JSON.stringify(requestHeaders) + " PT: " + proofType)
  if (requestHeaders.Proof != undefined) {
    switch (proofType) {
      case "SMT":
        let smtProof = JSON.parse(proof)
        return verifySMTProof(smtProof, 0);
      case "NoSQLScan":
      case "NoSQL":
        let nosqlProof = JSON.parse(proof)
        return verifyNoSQLProof(proofType, nosqlProof, path, 0, body, 0);
    }
  } else {
    // console.log("requested proof, but no proof in response");
    return true;
  }
  return false;
}

async function verifySMTProof(proofObject, verbose) {
  let proof = deserializeProof(proofObject.Proof);
  return await checkSMTProof(proof,
    hexStringToByte(proofObject.KeyHash),
    hexStringToByte(proofObject.TxHash),
    hexStringToByte(proofObject.MerkleRoot),
    verbose);
}

async function verifyNoSQLProof(proofType, proofObject, path, contentLen, body, verbose) {
  let wolkUrlObject = parseWolkPath(path)
  if (proofObject.Collection != wolkUrlObject.collection) {
    console.log("Collection Mismatch", proofObject.Collection, wolkUrlObject.collection)
    return false;
  }
  let defaultHashes = await computeDefaultHashes(160);

  if (proofType == "NoSQLScan") {
    if ( proofObject.ScanProofs != undefined ) {
      console.log("NoSQLScan with ScanProofs")
      for (var i = 0; i < proofObject.ScanProofs.length; i++) {
        let p = proofObject.ScanProofs[i];
        let prf = deserializeProof(p.KeyProof);
        if (proofObject.Collection.length == 0) {
          // TODO: add await
          let smtCheck = await checkSMTProof(prf, prf.Key, hexStringToByte(p.TxHash), hexStringToByte(proofObject.SystemMerkleRoot), defaultHashes, verbose)
          if (!smtCheck) {
            return false;
          }
        } else {
          let smtCheck = checkSMTProof(prf, prf.Key, hexStringToByte(p.TxHash), hexStringToByte(proofObject.CollectionMerkleRoot), defaultHashes, verbose)
          if (!smtCheck) {
            return false;
          }
        }
      }
    } else {
      if (proofObject.Collection.length == 0) { //listCollections
        if ( ! ( proofObject.SystemChunkHash === "0x44a00e7ae0499c6e377b95709f3851843976ea6c2a86eff46adfe15608d22005") ) {
          return false;
        }
        let prf = deserializeProof(proofObject.SystemProof)
        let smtCheck = checkSMTProof(prf, prf.Key, hexStringToByte(proofObject.SystemChunkHash), hexStringToByte(proofObject.KeyMerkleRoot), defaultHashes, verbose)
        if (!smtCheck) {
          return false;
        }
        console.log("NoSQLScan SUCCESS", "ch", proofObject.SystemChunkHash, "mr", proofObject.KeyMerkleRoot)
      } else { //scanCollection
        if ( ! ( proofObject.CollectionChunkHash === "0x44a00e7ae0499c6e377b95709f3851843976ea6c2a86eff46adfe15608d22005" ) ) {
          return false;
        }
      }
    }
    console.log("NoSQLScan Proof Verified")
    return true;
  }

  if (proofObject.Key != wolkUrlObject.path) {
    console.log(proofObject)
    console.log("Key Mismatch: [", proofObject.Key, "] key: [", wolkUrlObject.path, "]")
    return false;
  }

  // Verify Proof Process: at proof.Key position (which is keyToAddress(Key) where key is supplied by user)
  //   there is proof.TxHash, which hashes up to proof.CollectionMerkleRoot
  //console.log("proofObject.KeyProof: " + JSON.stringify(proofObject) )
  let keyProof = deserializeProof(proofObject.KeyProof);
  let keyHash = await keyToAddress(wolkUrlObject.path);
  if (!equalBytes(keyHash, keyProof.Key)) {
    console.log("Key Mismatch", keyProof.Key, "keyHash", keyHash, "path", wolkUrlObject.path);
    return false;
  }

  // position: keyhash, txhash = value hashes up to merkleroot
  let proof_txHash = hexStringToByte(proofObject.TxHash);
  let proof_collectionMerkleRoot = hexStringToByte(proofObject.CollectionMerkleRoot);
  let keyProofCheck = await checkSMTProof(keyProof, keyHash, proof_txHash, proof_collectionMerkleRoot, defaultHashes, verbose)
  if (!keyProofCheck) {
    console.log("KeyProof Verification FAILED");
    return false;
  }
  console.log("KeyProofCheck: ", keyProofCheck);
  if (proofObject.Tx == undefined) {
    console.log("Tx not supplied")
    return false;
  }

  // check that txhash matches provided tx ... TODO: check that pr.TxHash matches the Tx returned ***in RLP encoded form***
  let txhash = hexStringToByte(proofObject.Tx.txhash);
  let owner = wolkUrlObject.owner;
  let txsigner = hexStringToByte(proofObject.Tx.signer);
  /*
    if (!equalBytes(txhash, proof_txHash)) {
      console.log("Tx mismatch", txhash, proof_txHash);
      return false;
    }
  let txAmount = proofObject.Tx.amount;
  let contentLength = contentLen;
  if (contentLength != txAmount) {
    console.log("Header:Content-Length Does not Match Tx.Amount", "Header:Content-Length", contentLength, "Tx.Amount", txAmount);
    return false;
  }
  // TODO: check that TxSigner is valid, matches owner
  if (!equalBytes(txsigner, owner)) {
    // TODO: check against writers list?
    console.log("Signer mismatch");
    return false
  }
  if (!verifyContentHash(proofObject.Tx.GenericHash, body)) {
    console.log("ContentHash Mismatch");
    return false;
  }
  */

  // Verify Proof Process: at proof.Key position (which is proof.CollectionHash=CollectionHash(owner, collection))
  //   there is proof.CollectionChunkHash, which hashes up to proof.KeyMerkleRoot
  let collectionProof = deserializeProof(proofObject.CollectionProof);
  let collHash = await collectionHash(owner, wolkUrlObject.collection)
  if (!equalBytes(collectionProof.Key, collHash)) {
    console.log("Collection Mismatch", collectionProof.Key, "collHash", collHash);
    return false;
  }

  let proof_collectionChunkHash = hexStringToByte(proofObject.CollectionChunkHash);
  let proof_keyMerkleRoot = hexStringToByte(proofObject.KeyMerkleRoot);
  let collectionCheck = await checkSMTProof(collectionProof, collHash, proof_collectionChunkHash, proof_keyMerkleRoot, defaultHashes, verbose)
  if ( ! collectionCheck ) {
    console.log("CollectionProof Verification FAILED", collectionProof);
    return false;
  }
  console.log("CollectionProof Verified");

  // Verify proof process: at proof.Key position (which is proof.SystemHash=CollectionHash(owner, SystemCollection))
  //   there is the value proof.SystemChunkHash, which hashes up to proof.KeyMerkleRoot
  let systemProof = deserializeProof(proofObject.SystemProof)
  let systemHash = await keyToAddress(owner)
  if (!equalBytes(systemProof.Key, systemHash)) {
    console.log("System Mismatch")
    return false;
  }
  let proof_systemChunkHash = hexStringToByte(proofObject.SystemChunkHash);
  //let proof_keyMerkleRoot = hexStringToByte(proofObject.KeyMerkleRoot);
  let ownerCheck = await checkSMTProof(systemProof, systemHash, proof_systemChunkHash, proof_keyMerkleRoot, defaultHashes, verbose)
  if ( ! ownerCheck ) {
    console.log("SystemProof FAILED")
    return false;
  }
  console.log("SystemProof Verified");
  return true;
}

async function keyToAddress(k) {
  return new Uint8Array(await Computehash(unicodeStringToTypedArray(k))).subarray(0, 20);
}

function unicodeStringToTypedArray(s) {
  var escstr = encodeURIComponent(s);
  var binstr = escstr.replace(/%([0-9A-F]{2})/g, function(match, p1) {
    return String.fromCharCode('0x' + p1);
  });
  var ua = new Uint8Array(binstr.length);
  Array.prototype.forEach.call(binstr, function(ch, i) {
    ua[i] = ch.charCodeAt(0);
  });
  return ua;
}

async function collectionHash(owner, collection) {
  let x = await Computehash(Uint8Array.of(...unicodeStringToTypedArray(owner), ...unicodeStringToTypedArray(collection)));
  return new Uint8Array(x.slice(0, 20))
}

// GenerateRandomKey generates a random key of length l
function GenerateRandomKey(length) {
  /*TODO:
	key := make([]byte, l)
	var total int
	for total < l {
		read, _ := rand.Read(key[total:])
		total += read
	}
	return key
  */
}

function parseWolkPath(wolkPath) {
  let urlParts = wolkPath.split("/");
  var owner = ""
  var collection = ""
  var key = ""
  var wolkUrlObject = {
    owner: "",
    collection: "",
    path: "",
  };
  /* i.e. owner/collection/key -- NOTE: Different from parseWolkUrl */

  if (urlParts.length > 0) {
      wolkUrlObject.owner = urlParts.shift();
  }
  if (urlParts.length > 0) {
      wolkUrlObject.collection = urlParts.shift();
  }
  if (urlParts.length > 0) {
      wolkUrlObject.path = urlParts.join("/")
  }
  return wolkUrlObject;
}

if (typeof window !== 'undefined' && typeof window.WOLK === 'undefined') {
  window.WOLK = WOLK;
}
module.exports = WOLK;
