const idb = require("idb");
const BucketCloudstore = 0
const BucketNoSQL = 1
const BucketSQL = 2

const DEFAULT_DEVELOPER_TRUST_LEVEL = 0 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time
const DEFAULT_USER_TRUST_LEVEL = 0 //0 = regular web behavior no proof checking | 1 = check ALL proofs | In between check that % of time

class WOLK {

  constructor() {
    this.provider = "https://cloud.wolk.com"; //DEFAULT PROVIDER
    this.developerTrustLevel = DEFAULT_DEVELOPER_TRUST_LEVEL
    this.userTrustLevel = DEFAULT_USER_TRUST_LEVEL
    this.setWolkTrustLevel()
  }

  async init() {
    this.dbPromise = idb.openDb('names', 1, upgradeDB => {
      upgradeDB.createObjectStore('names', {
        keyPath: 'name'
      });
    });

    return await this.dbPromise.then( async db => {
      return await this.loadAccounts();
    })
    .catch(err => {
      console.log("WOLK: loadAccount error: " + err)
      throw new Error("Error loading accounts: " + err)
    });
  }

  setProvider(provider) {
    this.provider = provider;
  }

  setDeveloperTrustLevel( level ) {
    this.developerTrustLevel = level
  }

  setUserTrustLevel( level ) {
    this.userTrustLevel = level
  }

  setWolkTrustLevel() {
    if( this.userTrustLevel > this.developerTrustLevel ) {
      this.wolkTrustLevel  = this.userTrustLevel
    } else {
      this.wolkTrustLevel  = this.developerTrustLevel
    }
  }

  determineProofValue() {
    let randProof = Math.random();
    if ( randProof <= this.wolkTrustLevel ) {
      return 1
    } else {
      return 0
    }
  }

  loadAccounts(callback) {
    var that = this
    var keyRangeValue = null;
    return this.dbPromise
      .then( db => {
        return db.transaction(['names'], 'readonly').objectStore('names').getAll();
      })
      .then( accounts => {
        //console.log("WOLK: retrieved names: " + accounts + " json: " + JSON.stringify(accounts));
        if (accounts.length > 0) {
          let account = accounts[accounts.length - 1]
          console.log("WOLK: last Account", account)
          var accountEcdsaPrivateKey = account.ecdsaPrivateKey
          if ( typeof account.ecdsaPrivateKey == 'object' ) {
          } else {
            accountEcdsaPrivateKey = JSON.parse(account.ecdsaPrivateKey)
          }
          return window.crypto.subtle.importKey(
            "jwk",
            accountEcdsaPrivateKey, {
              name: "ECDSA",
              namedCurve: "P-256",
            },
            false,
            ["sign"])
            .then(function(ecdsaKey) {
              that.ecdsaKey = {}
              that.ecdsaKey.privateKey = ecdsaKey;
              that.ecdsaPublicKey = account.ecdsaPublicKey
              console.log("WOLK: loadAccounts", ecdsaKey)
              var accountRsaPrivateKey = account.rsaPrivateKey
              if ( typeof account.rsaPrivateKey == 'object' ) {
              } else {
                accountRsaPrivateKey = JSON.parse(account.rsaPrivateKey)
              }
              return window.crypto.subtle.importKey(
                  "jwk",
                  accountRsaPrivateKey, {
                    name: "RSASSA-PKCS1-v1_5",
                    hash: {  name: "SHA-256" },
                  },
                  false,
                  ["sign"]
                )
                .then(function(rsaKey) {
                  that.rsaKey = rsaKey;
                  console.log("Loaded")
                  return
                })
                .catch(function(err) {
                  console.error(err)
                });
            })
            .catch(function(err) {
              console.error(err)
            });
          }
          console.log('# of Accounts', accounts.length);
        })
        .catch( err => {
          console.log(err)
        })
  }

  async createAccount(name, callback) {
    let that = this
    name = name.toLowerCase();
    console.log("createAccount", name)
    // generate + export RSA Key
    await window.crypto.subtle.generateKey({
      name: "RSASSA-PKCS1-v1_5",
      modulusLength: 2048,
      publicExponent: new Uint8Array([0x01, 0x00, 0x01]),
      hash: {
        name: "SHA-256"
      }
    }, true, ["sign", "verify"])
    .then((rsaKey) => {
      var account = {}
      return window.crypto.subtle.exportKey("jwk", rsaKey.privateKey)
            .then(function(rsaPrivateKey) {
                account.rsaPrivateKey = rsaPrivateKey
                return window.crypto.subtle.exportKey("jwk", rsaKey.publicKey);
            })
            .then(function(rsaPublicKey) {
                account.rsaPrivateKey = rsaPublicKey
                return window.crypto.subtle.generateKey({ name: "ECDSA", namedCurve: "P-256"}, true, ["sign", "verify"])
            })
            .then(function(ecdsaKey) {
                return window.crypto.subtle.exportKey("jwk", ecdsaKey.privateKey)
                        .then(function(ecdsaPrivateKey) {
                            account.ecdsaPrivateKey = ecdsaPrivateKey
                            return window.crypto.subtle.exportKey("jwk", ecdsaKey.publicKey)
                        })
                        .then(function(ecdsaPublicKey) {
                            account.ecdsaPublicKey = ecdsaPublicKey
                            account.name = name
                            console.log(account);
                            return that.dbPromise
                              .then( db => {
                                return db.transaction(['names'], 'readwrite').objectStore('names').put(account);
                              })
                              .then( () => {
                                that.ecdsaPublicKey = JSON.stringify(account.ecdsaPublicKey)
                                that.ecdsaPrivateKey = account.ecdsaPrivateKey
                                that.ecdsaKey = ecdsaKey
                                return that.setName(name, JSON.stringify(account.rsaPublicKey), callback)
                              })
                              .catch( err => {
                                account = {}
                                console.error("ERROR sending name: " + err)
                              })
                      })
                      .catch( err => {
                        account = {}
                        console.error("ERROR setting up keys: " + err)
                      })
          })
          .catch( err => {
            account = {}
            console.err("ERROR parsing ecdsaKey: " + err)
          });
    })
    .catch( err => {
      account = {}
      console.err("ERROR generating account: " + err)
    });
  }

  setName(name, rsaPublicKey, callback) {
    return this.post(name, {
      "name": name,
      "rsaPublicKey": btoa(rsaPublicKey)
    }, {}, callback);
  }

  // Blockchain operations
  getBlock(blockNumber, callback) {
    return this.get(joinPath('wolk', 'block', blockNumber.toString(10)), {}, callback)
            .catch( (err) => { throw err; });
  }
  getNode(nodeNumber, callback) {
    return this.get(joinPath('wolk', 'node', nodeNumber.toString(10)), {}, callback)
            .catch( (err) => { throw err; });
  }
  updateNode(nodeNumber, nodeInfo, callback) {
    return this.patch(joinPath('wolk', 'node', nodeNumber.toString(10)), nodeInfo, {}, callback)
            .catch( (err) => { throw err; });
  }
  getPeers(nodeNumber, callback) {
    return this.get(joinPath('wolk', 'info'), {}, callback)
            .catch( (err) => { throw err; });
  }
  getName(name, callback) {
    return this.get(joinPath('wolk', 'name', name), {
      "Proof": this.determineProofValue()
      }, callback)
      .catch( (err) => { throw err; });
  }
  getAccount(name, callback) {
    return this.get(joinPath('wolk', 'account', name), {
      "Proof": this.determineProofValue()
      }, callback)
      .catch( (err) => { throw err; });
  }
  getLatestBlockNumber(callback) {
    return this.get(joinPath('wolk', 'block', 'latest'), {}, callback)
          .catch( (err) => { throw err; });
  }
  getTransaction(txhash, callback) {
    return this.get(joinPath('wolk', 'tx', txhash), {}, callback)
          .catch( (err) => { throw err; });
  }
  transfer(recipient, amount, callback) {
    return this.post(joinPath('wolk', 'transfer'), {
      "recipient": recipient,
      "amount": amount
    }, {}, callback);
  }
  // NoSQL
  createCollection(owner, collection, callback) {
    return this.post(joinPath(owner, collection), JSON.stringify({
      "name": collection,
      "bucketType": BucketNoSQL,
      "requesterPays": 0,
      "encryption": "none"
    }), {}, callback);
  }
  listCollections(owner, callback) {
    return this.get(owner, { "Proof": this.determineProofValue() }, callback)
          .catch( (err) => { throw err; });
  }
  listCollection(owner, collection, callback) {
    return this.get(joinPath(owner, collection), {
      "Proof": this.determineProofValue()
    }, callback);
  }
  deleteCollection(owner, collection, callback) {
    return this.delete(joinPath(owner, collection), {}, callback);
  }
  scanCollection(owner, collection, hdrs, callback) {
    return this.get(joinPath(owner, collection), { "Proof": this.determineProofValue() }, callback)
              .catch( (err) => { throw err; });
  }
  setKey(owner, collection, key, value, callback) {
    return this.put(joinPath(owner, collection, key), value, {}, callback);
  }
  getKey(owner, collection, key, hdrs, callback) {
    return this.get(joinPath(owner, collection, key), {
      "Proof": this.determineProofValue()
      }, callback)
      .catch( (err) => { throw err; });
  }
  deleteKey(owner, collection, key, callback) {
    return this.delete(joinPath(owner, collection, key), {}, callback);
  }

  // SQL
  createDatabase(owner, database, options, callback) {
    return this.post(joinPath(owner, database), JSON.stringify({
      "name": database,
      "bucketType": BucketSQL,
      "requesterPays": 0,
      "encryption": "none"
    }), options, {}, callback);
  }
  listDatabases(owner, callback) {
    return this.get(joinPath(owner), {
      "Proof": this.determineProofValue()
      }, callback)
      .catch( (err) => { throw err; });
  }
  deleteDatabase(owner, database, callback) {
    return this.delete(joinPath(owner, collection), {}, callback);
  }
  executeSQL(owner, database, sql, hdrs, callback) {
    return this.post(joinPath(owner, database, sql), sql, hdrs, callback);
  }

  // NOT EXPOSED
  async get(path, hdrs, callback) {
    return this.wrequest("GET", path, null, hdrs, callback);
  }

  async post(path, body, hdrs, callback) {
    return this.wrequest("POST", path, body, hdrs, callback);
  }

  async put(path, body, hdrs, callback) {
    return this.wrequest("PUT", path, body, hdrs, callback);
  }

  async patch(path, body, hdrs, callback) {
    return this.wrequest("PATCH", path, body, hdrs, callback);
  }

  async delete(path, hdrs, callback) {
    return this.wrequest("DELETE", path, null, hdrs, callback);
  }

  sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
  }

  async waitForTx(txhash, callback) {
    let done = false;
    let calls = 0;
    while (!done) {
      this.getTransaction(txhash, function(err, body, verified) {
        if (err) {
          //console.log("err:", err);
        } else {
          try {
            let tx = JSON.parse(body)
            if (tx.BlockNumber != undefined) {
              done = true;
              return callback(null, tx);
            }
          } catch (e) {

          }
        }
      })
      await this.sleep(500);
      if (calls++ > 15) {
        done = true;
        return callback(new Error('timeout, tx not included'), null);
      }
    }
  }
  async wrequest(method, path, body, hdrs, callback) {
    let enc = new TextEncoder("utf-8")
    let payloadBytes = method + "/" + path;
    if ( (method == "POST") || (method == "PATCH") ) {
      if (typeof body === "object") {
        body = JSON.stringify(body);
      }
      if ( body != null ) {
        payloadBytes += body;
      }
    }
    let payload = enc.encode(payloadBytes);
    let that = this

    var requestOptions = {
      method: method,
      headers: {
        "Requester": that.ecdsaPublicKey,
        "Msg": payloadBytes
      },
      body: body
    }

    if( typeof callback === 'function' ) {
      window.crypto.subtle.sign({ name: "ECDSA", hash: { name: "SHA-256" } }, this.ecdsaKey.privateKey, payload)
      .then(function(signature) {
        requestOptions.headers.Sig = buf2hex(signature);
        var url = that.provider + "/" + path
        fetch(url, requestOptions)
          .then(function(response) {
            if (response.ok) {
              return response.text()
            } else {
              console.log("ERROR:", response.status, response.statusText)
              // TODO: use response.status
              return callback(new Error(response.text()), null, false);
            }
          })
          .then(function(body) {
            let verified = true; // TODO: verifyHeaders(wolkUrlObject, hdrs, result.headers, body)
            return callback(null, body, verified);
          })
          .catch(function(err) {
            console.log(err)
            callback(err, null, false)
          })
      })
      .catch(function(err) {
        console.error(err);
        callback(err, null, false)
      });
    } else {
      console.log("WOLK: Promise route")
      return window.crypto.subtle.sign({ name: "ECDSA", hash: { name: "SHA-256" } }, this.ecdsaKey.privateKey, payload)
      .then(function(signature) {
        requestOptions.headers.Sig = buf2hex(signature);
        var url = that.provider + "/" + path
        return fetch(url, requestOptions)
          .then(function(response) {
            if (response.ok) {
              return response.text()
            } else {
              console.log("ERROR:", response.status, response.statusText)
              // TODO: use response.status
              throw new Error(response);
              //return response
            }
          })
          .then(function(body) {
            let verified = true; // TODO: verifyHeaders(wolkUrlObject, hdrs, result.headers, body)
            return body;
          })
          .catch(function(err) {
            console.log(err)
            throw err;
              // callback(err, null, false)
          })
      })
      .catch(function(err) {
        console.error(err);
        throw err;
      });
    }
  }
}

function buf2hex(buffer) { // buffer is an ArrayBuffer
  return Array.prototype.map.call(new Uint8Array(buffer), x => ('00' + x.toString(16)).slice(-2)).join('');
}

joinPath = (...args) => {
  return args.map((part, i) => {
    if (i === 0) {
      return part.trim().replace(/[\/]*$/g, '')
    } else {
      return part.trim().replace(/(^[\/]*|[\/]*$)/g, '')
    }
  }).filter(x => x.length).join('/')
}

function sum(a, b) {
  return a + b;
}

if (typeof window !== 'undefined' && typeof window.WOLK === 'undefined') {
    window.WOLK = WOLK;
}

module.exports = WOLK;
