var assert = chai.assert;

function randomString(n) {
  return Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 5);
}

let OWNER = "account" + randomString(5);
let DATABASE = "db" + randomString(5);
let TABLE = "tbl" + randomString(5);
let TEST_MAX_MS = 10000;

describe('Account Creation', function() {
  this.timeout(TEST_MAX_MS);

  it('getName(' + OWNER + ') not found', function(done) {
    wolkjs.getName(OWNER)
      .catch(function(err) {
        assert(isNotFound(err))
        done();
      });
  });

  it('createAccount(' + OWNER + ')', function(done) {
    wolkjs.createAccount(OWNER, function(err, txhash) {
      wolkjs.waitForTx(txhash, function(err, tx) {
        if (err) {
          throw err;
        }
        console.log("===> setName TX: ", txhash, tx)
        done();
      })
    });
  });

  it('getName(' + OWNER + ') found', function(done) {
    wolkjs.getName(OWNER)
      .then(function(result) {
        console.log(result)
        done();
      });
  });
});

describe('Create Database', function() {
  this.timeout(15000);
  let VAL = "val" + randomString(5);

  it('listDatabases(' + OWNER + ') [should be empty]', function(done) {
    wolkjs.listDatabases(OWNER, {})
      .then(function(result) {
        let databases = JSON.parse(result);
        console.log("databases:", databases)
        // TODO: assert(databases.data == null) [currently ownerchunk has gibberish]
        done();
      })
  });

  it('createDatabase(' + OWNER + ', ' + DATABASE + ')', function(done) {
    wolkjs.createDatabase(OWNER, DATABASE, {})
      .then(function(txhash) {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash)
          .then(function(tx) {
            console.log("===> createDatabase TX: ", txhash, tx)
            done();
          });
      })
  });

  it('listDatabases(' + OWNER + ')', function(done) {
    wolkjs.listDatabases(OWNER, {})
      .then(function(result) {
        let databases = JSON.parse(result);
        console.log("databases:", databases.data)
        assert(databases.data.length > 0);
        done();
      })
  });
});

describe('Create Table', function() {
  this.timeout(TEST_MAX_MS);
  let columns = [{
    "columnname": "person_id",
    "indextype": "BPLUS",
    "columntype": "INTEGER",
    "primary": 1
  }, {
    "columnname": "name",
    "indextype": "BPLUS",
    "columntype": "STRING"
  }]

  it('list tables [should be empty]', function(done) {
    wolkjs.listTables(OWNER, DATABASE, {})
      .then(function(result) {
        console.log("list tables:", result)
        let res = JSON.parse(result);
        assert(res.data == null);
        done();
      })
  });

  // wcloud sql wolk://OWNER/DATABASE "create table personnel (person_id int primary key, name string)"
  it('create table ' + TABLE, function(done) {
    wolkjs.createTable(OWNER, DATABASE, TABLE, columns, {})
      .then(function(txhash) {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash)
          .then(function(tx) {
            console.log("===> Create Table TX: ", txhash, tx)
            done();
          });
      })
  });

  it('list tables', function(done) {
    wolkjs.listTables(OWNER, DATABASE, {})
      .then(function(result) {
        console.log("list tables:", result)
        let res = JSON.parse(result);
        assert(res.data.length == 1);
        done();
      })
  });

  it('describe table', function(done) {
    wolkjs.describeTable(OWNER, DATABASE, TABLE, {})
      .then(function(result) {
        // {"data":[{"ColumnName":"name","ColumnType":"STRING","IndexType":"BPLUS","Primary":0},{"ColumnName":"person_id","ColumnType":"INTEGER","IndexType":"BPLUS","Primary":1}]}
        let res = JSON.parse(result);
        console.log("describe table:", res)
        assert(res.data.length == 2);
        done();
      })
  });

});

describe('Mutate/Read', function() {
  this.timeout(TEST_MAX_MS);
  let PERSON_ID = Math.floor(Math.random() * Math.floor(10000));
  let NAME = "name" + randomString(5);

  let sql_insert = "insert into " + TABLE + " (person_id, name) values (" + PERSON_ID + ", '" + NAME + "')";
  let sql_select = "select * from " + TABLE;

  it(sql_select + " [should be empty]", function(done) {
    console.log(sql_select)
    wolkjs.readSQL(OWNER, DATABASE, sql_select, {})
      .then(function(result) {
        let res = JSON.parse(result);
        assert(res.data == null);
        done();
      });
  });

  it(sql_insert, function(done) {
    console.log(sql_insert)
    wolkjs.executeSQL(OWNER, DATABASE, sql_insert, {})
      .then(function(txhash) {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash)
          .then(function(tx) {
            console.log("===> INSERT TX: ", txhash, tx)
            done();
          })
      })
  })

  it(sql_select + " [should have one record]", function(done) {
    console.log(sql_select)
    wolkjs.readSQL(OWNER, DATABASE, sql_select, {})
      .then(function(result) {
        let res = JSON.parse(result);
        assert(res.data.length == 1);
        let rec = res.data[0];
        console.log("RES:", res.data[0])
        assert(rec.name == NAME)
        assert(rec.person_id == PERSON_ID)
        done();
      });
  });
});

describe('Drop table', function() {
  this.timeout(TEST_MAX_MS);
  it('drop table ' + TABLE, function(done) {
    wolkjs.dropTable(OWNER, DATABASE, TABLE, {})
      .then(function(txhash) {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash)
          .then(function(tx) {
            console.log("===> drop table TX:", txhash, tx)
            done();
          });
      })
  });

  it('listTables', function(done) {
    wolkjs.listTables(OWNER, DATABASE, {})
      .then(function(result) {
        let tables = JSON.parse(result);
        console.log("tables:", tables);
        assert(tables.data == null)
        done();
      })
  });
});

describe('Delete database', function() {
  this.timeout(TEST_MAX_MS);
  it('deleteDatabase', function(done) {
    wolkjs.deleteDatabase(OWNER, DATABASE)
      .then(function(txhash) {
        assert(txhash.length >= 64);
        wolkjs.waitForTx(txhash)
          .then(function(tx) {
            console.log("===> delete database TX:", txhash, tx)
            done();
          });
      })
  });

  it('listDatabases(' + OWNER + ')', function(done) {
    wolkjs.listDatabases(OWNER, {})
      .then(function(result) {
        let databases = JSON.parse(result);
        console.log("databases:", databases)
        // TODO: assert(databases.data == null) [currently ownerchunk has gibberish]
        done();
      })
  })

});
